from io import StringIO
import z3
import jz3.src.run_solvers as run_solvers
import warnings


# child class to write push and pop to SMT2 file
class Solver(z3.Solver):
    def __init__(self, benchmark_mode=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__start_recording = False
        self.__history = []
        self.__assertions = []
        self.__global_constraints = z3.BoolVal(True)
        self.__smt_str = ""
        self.__condition_var_assignment_model = None
        self.__latest_solvers_results = None
        self.__benchmark_mode = benchmark_mode
        self.__variables = set()

    def __getattribute__(self, name):
        _allowed_methods = ['add', 'add_global_constraints', 'add_conditional_constraint',
                            'check_conditional_constraints', 'check', 'push', 'pop',
                            'start_recording', 'generate_smtlib', '_allowed_methods',
                            'ctx', 'solver', 'set', 'assert_exprs', 'to_smt2', 'assertions',
                            'get_condition_var_assignment_model',
                            'get_latest_solvers_results']
        if name.startswith('_') or name in _allowed_methods:  # intentionally accessing a private variable
            return object.__getattribute__(self, name)
        else:
            warnings.warn(f"Method '{name}' is called.\n "
                          f"But this method might not be recorded to smt2 file and might incur potential logic errors"
                          f"Please use only the methods defined in Solver2SMT.\n"
                          f"If this is intentional, modified the _allowed_methods above")
            return super().__getattribute__(name)

    def add_global_constraints(self, *constraints):
        """
        Sets global constraints that encodes rules/constraints for the condition variables.
        :param constraints: A list of Z3 constraints that define global conditions.
        """
        self.__global_constraints = z3.And(self.__global_constraints, *constraints)

    def add(self, *args):
        # self._conditional_constraints.append((args,condition))
        if self.__start_recording:
            for arg in args:
                self.__history.append(("add", str(arg.sexpr())))
        super().add(*args)

    def add_conditional_constraint(self, *args, condition=z3.BoolVal(True)):
        if condition is None:
            condition = z3.BoolVal(True)
        for conditional_constraint in args:
            self.__assertions.append((conditional_constraint, condition))
            self.__variables.add(condition)
        s = z3.Solver()
        s.add(self.__global_constraints)

        if s.check() != z3.sat:
            raise "There is no way to satisfy all condition variables provided under global constraint"

    def check_conditional_constraints(self, *args, condition=z3.BoolVal(True),max_count=5):
        s = z3.Solver()  # no smt file recording required
        s.add(self.__global_constraints)

        # temporarily add the constraint and conditional constraint to be checked.
        if args:  # append the checked condition
            self.__assertions.append((args, condition))

        if s.check() == z3.sat:
            # possible combination of condition variables
            model = s.model()

            solver_with_conditional_constraint = Solver()

            # add corresponding conditional constraints and try to solve
            for (conditional_constraint, condition) in self.__assertions:
                if condition == z3.BoolVal(True) or model.eval(condition):
                    if self.__start_recording:
                        self.__history.append(("add", str(conditional_constraint.sexpr())))
                    solver_with_conditional_constraint.add(conditional_constraint)

            # Don't really record the smt files
            solver_with_conditional_constraint.start_recording()
            result = solver_with_conditional_constraint.check()

            self.__condition_var_assignment_model = [model]

            # Only launch multiple solvers when in benchmark mode
            if self.__benchmark_mode:
                self.__condition_var_assignment_model = []
                self.__latest_solvers_results = []

                # find different combinations
                opt = z3.Optimize()
                opt.add(self.__global_constraints)
                dist = 1
                count = 0
                while (count < max_count) and (dist > 0):
                    solver_with_conditional_constraint = Solver()

                    # add corresponding conditional constraints and try to solve
                    for (conditional_constraint, condition) in self.__assertions:
                        if model.eval(condition):
                            if self.__start_recording:
                                self.__history.append(("add", str(conditional_constraint.sexpr())))
                            solver_with_conditional_constraint.add(conditional_constraint)

                    # append the combination to the results
                    # solver_with_conditional_constraint.start_recording()
                    result = solver_with_conditional_constraint.check()
                    single_condition_smt_str = solver_with_conditional_constraint.to_smt2()
                    print(single_condition_smt_str)
                    self.__latest_solvers_results.append(run_solvers.run_solvers(smt2_str=single_condition_smt_str,verbose=True))
                    self.__condition_var_assignment_model.append(model)
                    min_hamdist = z3.Int("min_hamdist")

                    opt.add(min_hamdist <= z3.Sum(
                        tuple(z3.If((var == bool(model[var])), 0, 1) for var in self.__variables)))

                    opt.maximize(min_hamdist)
                    opt.check()
                    model = opt.model()
                    dist = model[min_hamdist].as_long()
                    count += 1

                # store smt file/str
                self.__smt_str = solver_with_conditional_constraint.generate_smtlib()

                with open("conditional_constraints.smt2", "w") as file: # TODO
                    file.write(self.__smt_str)

                # launch multiple solvers and store resutls

            # pop the temporarily added conditional constraints
            if args:
                self.__assertions.pop()

            if self.__start_recording:
                self.__history.append(("result", str(solver_with_conditional_constraint.check(*args))))
            return result
        else:
            raise ("Impossible to find any way of building constraints"
                   "The conditional constraints are not satisfiable under global constraints ")

    def push(self):
        if self.__start_recording:
            self.__history.append(("push", None))
        super().push()

    def pop(self, *args, **kwargs):
        if self.__start_recording:
            self.__history.append(("pop", None))
        super().pop(*args, **kwargs)

    def check(self, *args, **kwargs):
        result = None
        if self.__start_recording:
            if args:
                for arg in args:
                    # Format the args in SMT-LIB syntax
                    self.__history.append(("push", ""))
                    self.__history.append(("add", str(arg.sexpr())))
                    self.__history.append(("check", ""))
                    result = super().check(*args)
                    self.__history.append(("result", result))
                    self.__history.append(("pop", ""))
            else:
                self.__history.append(("check", ""))
                result = super().check(*args)
                self.__history.append(("result", result))
            return result
        else:
            return super().check(*args)

    def start_recording(self):
        self.__start_recording = True
        self.check()
        initial_state = self.to_smt2()
        # Remove the last "(check-sat)" from the initial state
        initial_state = initial_state.rsplit("(check-sat)", 1)[0]
        self.__history.append(("initial_state", initial_state))

    def generate_smtlib(self):
        output = StringIO()
        output.write(f"(set-logic QF_LIA)\n")
        for operation in self.__history:
            op, args = operation
            if op == "initial_state":
                output.write(args)
            elif op == "add":
                output.write(f"(assert {args})\n")
            elif op in ["push", "pop"]:
                output.write(f"({op} 1)\n")
            elif op == "check":
                output.write("(check-sat)\n")
            elif op == "result":
                output.write(f"; Result: {args}\n")

        smt_str = output.getvalue()
        output.close()
        return smt_str

    def get_condition_var_assignment_model(self):
        return self.__condition_var_assignment_model

    def get_latest_solvers_results(self):
        return self.__latest_solvers_results


def solver_demo():
    solver = Solver(benchmark_mode=True)

    x = z3.Int('x')
    y = z3.Int('y')

    solver.add(x > 0)
    solver.add(y > 0)

    condition1 = z3.Bool('condition1')
    condition2 = z3.Bool('condition2')

    solver.add_global_constraints(z3.Or(condition1, condition2))
    solver.add_global_constraints(z3.Distinct(condition1,condition2))

    solver.add_conditional_constraint(x < 5, condition=condition1)
    solver.add_conditional_constraint(x > 5, condition=condition2)

    solver.start_recording()
    result = solver.check_conditional_constraints()
    print(result)

    # Access the recorded combinations and performance results
    print("Condition Variable Assignment Models:")
    print(solver.get_condition_var_assignment_model())
    print("Latest Solvers Results:")
    print(solver.get_latest_solvers_results())


def optimizer_test():
    # Create variables
    x = z3.Int('x')
    y = z3.Int('y')
    z = z3.Int('z')

    # Create conditions
    cond1 = z3.Bool('cond1')
    cond2 = z3.Bool('cond2')
    cond3 = z3.Bool('cond3')

    # Create constraints
    constraints = [
        z3.Implies(cond1, x > 0),
        z3.Implies(cond2, y > 0),
        z3.Implies(cond3, z > 0)
    ]

    # Create optimizer
    opt = z3.Optimize()

    # Add constraints to the optimizer
    for constraint in constraints:
        opt.add(constraint)

    # Add conditions to the optimizer
    opt.add(z3.Or(cond1, cond2, cond3))

    # First combination (maximize)
    opt.push()
    opt.maximize(z3.Sum([z3.If(cond, 1, 0) for cond in [cond1, cond2, cond3]]))
    opt.check()
    combination1 = opt.model()
    opt.pop()

    # Second combination (minimize)
    opt.push()
    opt.minimize(z3.Sum([z3.If(cond, 1, 0) for cond in [cond1, cond2, cond3]]))
    combination2 = opt.model()
    opt.pop()

    # Print the combinations
    print("Combination 1:")
    print(combination1)
    print("Combination 2:")
    print(combination2)

if __name__ == '__main__':
    solver_demo()
    # optimizer_test()



# install openyxl
# install xlwt
import numpy as np
import pandas as pd
import glob
from displayfunction import display
from pandas import ExcelWriter




# TODO: Adjust penalty weight
penalty_weight = 0  # seconds
excel_file_path = 'archive/test.xlsx'

def create_table():
    """
    Creates a table containing the average time of each constraint
    :return: the table: 2D list
    """
    table = [[b1, b2, b3, b4, b5] for b1 in (True, False)
             for b2 in (True, False)
             for b3 in (True, False)
             for b4 in (True, False) if not (b2 and b4)
             for b5 in (True, False)]

    for ele in table:
        b1, b2, b3, b4, b5 = ele
        for suffix in ['full_time.txt', 'holes_time.txt']:
            s = ['argyle-', 'classic-'][b1] + ['PbEq-', 'distinct-'][b2] + ['inorder-', 'percol-'][b3] \
                + ['is_num-', 'is_bool-'][b4] + ['no_prefill-', 'prefill-'][b5] + suffix
            try:
                with open('time-record/' + s) as f:
                    lines = f.read().split('\n')[:-1]  # strip the last whitespace
            except FileNotFoundError:
                lines = []
            # ['36.175875663757324,1', '36.49494504928589,1']
            data = [line.split(',') for line in lines]
            runtimes = [float(datum[0]) for datum in data]
            timeouts = [int(datum[1]) for datum in data]
            average_time = np.average(runtimes)
            num_rows = len(data)
            # percentage of timeouts:
            timeout_perc = np.average([(0 if timeout == 0 else 1) for timeout in timeouts])
            # average timeout
            timeout_avg = np.average([timeout for timeout in timeouts if timeout != 0])

            yield [b1, b2, b3, b4, b5, suffix == 'full_time.txt',
                   average_time, num_rows, timeout_perc, timeout_avg]
    return table

def export_to_excel():
    table = create_table()
    # export to excel
    df = pd.DataFrame(table, columns=['classic', 'distinct', 'per_col', 'no_num', 'prefill', 'generating full grid',
                                      'average time', 'number of rows', 'percentage with any timeouts', 'avg nr of timeouts'])
    time_rec = df['average time']
    solve_time = np.array(time_rec[::2])
    gen_time = np.array(time_rec[1::2])
    assert len(solve_time) == len(gen_time), "some full/holes sudokus are missing"
    df['full/holes ratio'] = [None if i % 2 == 0 else solve_time[i//2]/gen_time[i//2] for i in range(len(time_rec))]
    # print(df)
    with ExcelWriter(excel_file_path) as writer:
        df.to_excel(writer,sheet_name="Sheet1")
        # worksheet = writer.sheets["Sheet1"]
        # print(type(worksheet))
        # worksheet.cell()
        # worksheet.set_column(1, 1, 18)

    print("Successfully genreated excel file at ", excel_file_path)


if __name__ == '__main__':
    pass






import ast
import matplotlib.pyplot as plt
import numpy as np


class ConstraintPlotter:
    def __init__(self, file_path):
        self.file_path = file_path
        self.parsed_data = self._parse_data()
        self.x_max = 5
        self.y_max = 5
        self.width = 10
        self.height = 5
        self.opacity = 0.6
        self.marker_size=None
        self.line_style = 'r--'
        self.grid = True

    def set_graph_properties(self, x_max=5, y_max=5, width=10, height=5, opacity=0.6, marker_size=None,
                             line_style='r--', grid=True):
        self.x_max = x_max
        self.y_max = y_max
        self.width = width
        self.height = height
        self.opacity = opacity
        self.marker_size = marker_size
        self.line_style = line_style
        self.grid = grid

    def _parse_data(self):
        parsed_data = []
        with open(self.file_path, 'r') as file:
            for line in file:
                try:
                    data_dict = ast.literal_eval(line.strip())
                    parsed_data.append(data_dict)
                except Exception as e:
                    print("Error while parsing the dictionary, but continuing...")
                    print(f"The line that incurred the issue is: ---->\n{e}\n<----- If nothing is printed in between, it's an empty space or new line encountered")
                    continue
        return parsed_data

    def plot_constraints_comparison(self, constraint_comparison_num, solvers=None, combined_plot=False, constraint_names=None):
        times_constraint_true = {}
        times_constraint_false = {}
        combined_times_true = []
        combined_times_false = []

        for entry in self.parsed_data:
            processed_keys = set()  # To keep track of processed keys
            for key in list(entry.keys()):
                if key not in processed_keys and isinstance(key, tuple):
                    # Create the complement key
                    complement_key = list(key)
                    complement_key[constraint_comparison_num] = not key[constraint_comparison_num]
                    complement_key = tuple(complement_key)
                    if complement_key in entry:
                        # Determine which is the True key and which is the False key
                        true_key, false_key = (key, complement_key) if key[constraint_comparison_num] else (
                        complement_key, key)

                        true_data = entry[true_key]
                        false_data = entry[false_key]

                        for solver_key in true_data:
                            if solver_key != 'smt_path' and (solvers is None or solver_key in solvers):
                                time_true = true_data[solver_key][0][0]
                                time_false = false_data[solver_key][0][0]

                                if solver_key not in times_constraint_true:
                                    times_constraint_true[solver_key] = []
                                if solver_key not in times_constraint_false:
                                    times_constraint_false[solver_key] = []

                                times_constraint_true[solver_key].append(time_true)
                                combined_times_true.append(time_true)

                                times_constraint_false[solver_key].append(time_false)
                                combined_times_false.append(time_false)

                        processed_keys.add(true_key)
                        processed_keys.add(false_key)

        if solvers is None:
            solvers = list(times_constraint_true.keys())

        num_plots = len(solvers) + 1 if combined_plot else len(solvers)
        fig, axs = plt.subplots(num_plots, 1, figsize=(self.width, self.height * num_plots))

        if num_plots == 1:
            axs = [axs]

        for i, solver_key in enumerate(solvers):
            if solver_key in times_constraint_true and solver_key in times_constraint_false:
                min_length = min(len(times_constraint_true[solver_key]), len(times_constraint_false[solver_key]))
                true_times = times_constraint_true[solver_key][:min_length]
                false_times = times_constraint_false[solver_key][:min_length]
                # adjusting the datapoints to be within or at the edge of the graph
                true_times = np.clip(true_times, 0, self.x_max)
                false_times = np.clip(false_times, 0, self.y_max)
                for ele in false_times:
                    if ele > self.y_max:
                        raise "AS"

                axs[i].scatter(true_times, false_times, alpha=self.opacity,s=self.marker_size)
                axs[i].plot([0, self.x_max], [0, self.y_max], self.line_style)
                axs[i].set_title(f'Time comparison for {solver_key}')

                # set x y label according to constraint name
                if constraint_names and len(constraint_names) > constraint_comparison_num:
                    constraint_name_true, constraint_name_false = constraint_names[constraint_comparison_num]
                    axs[i].set_xlabel(f'Time when {constraint_name_true}')
                    axs[i].set_ylabel(f'Time when {constraint_name_false}')
                else:
                    axs[i].set_xlabel(f'Time when constraint {constraint_comparison_num} is True')
                    axs[i].set_ylabel(f'Time when constraint {constraint_comparison_num} is False')

                axs[i].set_xlim([0, self.x_max])
                axs[i].set_ylim([0, self.y_max])
                axs[i].grid(self.grid)

        if combined_plot:
            min_length_combined = min(len(combined_times_true), len(combined_times_false))
            # assert len(combined_times_false)==len(combined_times_true), f"There are unequal cases of conditions {len(combined_times_true)}, {len(combined_times_false)}"
            combined_true = combined_times_true[:min_length_combined]
            combined_false = combined_times_false[:min_length_combined]

            axs[-1].scatter(combined_true, combined_false, alpha=self.opacity,s=self.marker_size, color='green')
            axs[-1].plot([0, self.x_max], [0, self.y_max], 'r--')
            axs[-1].set_title('Combined Time Comparison for All Solvers')

            # set x y label according to constraint name
            if constraint_names and len(constraint_names) > constraint_comparison_num:
                constraint_name_true, constraint_name_false = constraint_names[constraint_comparison_num]
                axs[-1].set_xlabel(f'Time when {constraint_name_true} (All Solvers)')
                axs[-1].set_ylabel(f'Time when {constraint_name_false} (All Solvers)')
            else:
                axs[-1].set_xlabel(f'Time when constraint {constraint_comparison_num} is True (All Solvers)')
                axs[-1].set_ylabel(f'Time when constraint {constraint_comparison_num} is False (All Solvers)')

            axs[-1].set_xlim([0, self.x_max])
            axs[-1].set_ylim([0, self.y_max])
            axs[-1].grid(self.grid)

        plt.tight_layout()
        plt.show()


if __name__ == '__main__':
    time_instances_file_path = '../../time-record/particular_hard_instance_time_record/argyle_time.txt'
    constraint_names = [("classic","argyle"),
                        ("distinct","PbEq"),
                        ("percol","inorder"),
                        ("is_bool","is_num"),
                        ("prefill","no_prefill"),
                        ("gen_time","solve_time")]
    plotter = ConstraintPlotter(time_instances_file_path)
    plotter.set_graph_properties(x_max=5,y_max=5)
    # plotter.plot_constraints_comparison(3, solvers=["z3", "cvc5"],  combined_plot=True, constraint_names=constraint_names)
    plotter.plot_constraints_comparison(1, solvers=["z3","cvc5"],  combined_plot=True, constraint_names=constraint_names)


# found a way to avoid these: just create a super class
## Functions that the classes share:
# obj.z3():
#   get the z3 version of the formula obj (to hand to the z3 solver)
# obj.compute(dict):
#   get a true/false value based on obj by looking up the variables in dict
# obj.predicates():
#   get the set of predicates occurring in obj
from abc import ABC, abstractmethod
import z3
from typing import Tuple, Dict, List


class Expression(ABC):
    """Abstract base class for all expressions."""

    @abstractmethod
    def get_predicate_name(self):
        """Returns the variable(s) name(s). involved in the expression"""
        pass

    @abstractmethod
    def to_z3_expr(self):
        """Convert to z3 expression."""
        pass

    @abstractmethod
    def evaluate_assigned_value(self, value_dct):
        """return the assigned value for this variable/predicate in value_dct"""
        pass


class BoolVal(Expression):
    def __init__(self, value):
        self.value = value

    def get_predicate_name(self):
        return {}

    def evaluate_assigned_value(self, _values):
        return self.value

    def to_z3_expr(self):
        return z3.BoolVal(self.value)


class Bool(Expression):
    """Wrapper for boolean variable."""

    def __init__(self, name):
        self.name = name

    def get_predicate_name(self):
        return self.name

    def to_z3_expr(self):
        return z3.Bool(self.name)

    def evaluate_assigned_value(self, value_dct):
        return value_dct[self.name]


class Int(Expression):
    """Wrapper for integer variable."""

    def __init__(self, name):
        self.name = name

    def get_predicate_name(self):
        return self.name

    def to_z3_expr(self):
        return z3.Int(self.name)

    def evaluate_assigned_value(self, value_dct):
        return value_dct[self.name]


class Not(Expression):
    """Wrapper for logical NOT operation."""

    def __init__(self, *args):
        self.args = args

    def get_predicate_name(self):
        names = set()
        for arg in self.args:
            names.update(*arg.get_predicate_name())
        return names

    def to_z3_expr(self):
        return z3.Not(self.arg.to_z3_expr())

    def evaluate_assigned_value(self, value_dct):
        return not self.arg.evaluate_assigned_value(value_dct)


class Or(Expression):
    """Wrapper for logical OR operation between multiple expressions."""

    def __init__(self, *args):
        self.args = args

    def get_predicate_name(self):
        names = set()
        for arg in self.args:
            names.update(*arg.get_predicate_name())
        return names

    def to_z3_expr(self):
        return z3.Or([arg.to_z3_expr() for arg in self.args])

    def evaluate_assigned_value(self, value_dct):
        return any(arg.evaluate_assigned_value(value_dct) for arg in self.args)


class And(Expression):
    """Wrapper for logical AND operation between multiple expressions."""

    def __init__(self, *args):
        self.args = args

    def get_predicate_name(self):
        names = set()
        for arg in self.args:
            names.update(*arg.get_predicate_name())
        return names

    def to_z3_expr(self):
        for arg in self.args:
            print(arg)
        return z3.And([arg.to_z3_expr() for arg in self.args])

    def evaluate_assigned_value(self, value_dct):
        return all(arg.evaluate_assigned_value(value_dct) for arg in self.args)


class Distinct(Expression):
    """...."""

    def __init__(self, *args):
        self.args = args

    def get_predicate_name(self):
        names = set()
        for arg in self.args:
            names.update(*arg.get_predicate_name())
        return names

    def to_z3_expr(self):
        return z3.Distinct([arg.to_z3_expr() for arg in self.args])

    def evaluate_assigned_value(self, value_dct):
        all_predicates = [arg.evaluate_assigned_value(value_dct) for arg in self.args]
        return len(set(all_predicates)) == len(all_predicates)



class PbEq(Expression):
    """...."""

    def __init__(self, expr_weights: List[Tuple[Expression, int]], equal_val: int):
        self.expr_weights = expr_weights
        self.equal_val = equal_val


    def get_predicate_name(self):
        # Implementation depends on specifics of PbEq.
        names = set()
        for name,_ in self.expr_weights:
            names.update(name.get_predicate_name())
        return names

    def to_z3_expr(self):
        # Implementation depends on specifics of PbEq.
        return z3.PbEq(self.expr_weights,self.equal_val)

    def evaluate_assigned_value(self, value_dct):
        sum_values = sum(expr.evaluate_assigned_value(value_dct) * weight for expr, weight in self.expr_weights)
        return sum_values==self.equal_val


class Const(Expression):
    """Wrapper for constant value."""

    def __init__(self, value):
        self.value = value

    def get_predicate_name(self):
        return set()

    def to_z3_expr(self):
        # Directly return the constant value as its Z3 expression equivalent
        return z3.Const(self.value)

    def evaluate_assigned_value(self, value_dct):
        # The evaluation of a constant is the constant itself
        return self.value


class Eq(Expression):

    def __init__(self, expr1, expr2):
        self.expr1 = expr1
        self.expr2 = expr2

    def get_predicate_name(self):
        return (self.expr1, self.expr2)

    def to_z3_expr(self):
        return z3.eq(self.expr1.to_z3_expr(), self.expr2.to_z3_expr())

    def evaluate_assigned_value(self, value_dct):
        return value_dct[self.expr1] == value_dct[self.expr2]

class PbLe(Expression):
    """...."""

    def __init__(self, expr_weights: List[Tuple[Expression, int]], equal_val: int):
        self.expr_weights = expr_weights
        self.equal_val = equal_val


    def get_predicate_name(self):
        # Implementation depends on specifics of PbEq.
        names = set()
        for name,_ in self.expr_weights:
            names.update(name.get_predicate_name())
        return names

    def to_z3_expr(self):
        # Implementation depends on specifics of PbEq.
        return z3.PbEq(self.expr_weights,self.equal_val)

    def evaluate_assigned_value(self, value_dct):
        sum_values = sum(expr.evaluate_assigned_value(value_dct) * weight for expr, weight in self.expr_weights)
        return sum_values<=self.equal_val

class Implies(Expression):
    """Represents a logical implication between two expressions."""

    def __init__(self, premise, conclusion):
        self.premise = premise
        self.conclusion = conclusion

    def get_predicate_name(self):
        return self.premise.get_predicate_name(), self.conclusion.get_predicate_name()

    def to_z3_expr(self):
        return z3.Implies(self.premise.to_z3_expr(), self.conclusion.to_z3_expr())

    def evaluate_assigned_value(self, value_dct):
        return not self.premise.evaluate_assigned_value(value_dct) or self.conclusion.evaluate_assigned_value(value_dct)


def true():
    return BoolVal(True)


def false():
    return BoolVal(False)


class Solver:
    def __init__(self):
        self.assertions: List[Tuple[Expression, Expression]] = []  # (conditional_constraint, condition)
        self.modelVariables = {}  # no_num, distinct, etc
        self.global_constraints = true()  # no_num and distinct cannot be true at the same time
        self.model = None

    def add(self, *args, condition: Expression = None) -> None:
        """
        Adds an expression as a constraint to the solver, optionally conditioned on another expression.

        Parameters:
            expression (Expression): The constraint to be added.
            condition (Expression, optional): The condition under which the constraint is applied.
        """

        if condition is None:
            condition = true()
        for conditional_constraint in args:
            # TODO: check the the constraints are validly typed @sj is this what you mean?
            self.assertions.append((conditional_constraint, condition))
        s = z3.Solver()
        s.add(self.global_constraints.to_z3_expr())
        for _, condition in self.assertions:
            # Check if the condition is an instance of your Expression class
            # and convert it to a Z3 expression if so. Otherwise, use it as is.
            if isinstance(condition, Expression):
                z3_condition = condition.to_z3_expr()
            else:
                z3_condition = condition
            s.add(z3_condition)
        if s.check() != z3.sat:
            raise "The conditions provided are not satisfiable"

    def check(self, *args, condition: Expression = true()):
        """
        Finds a satisfiable set of conditional variables that satisfies the global constraints
        and an optional specific condition. It then applies corresponding constraints based on this set
        and checks for overall satisfiability.

        """
        s = z3.Solver()
        s.add(self.global_constraints.to_z3_expr())
        # unconditional constraint

        # adding all conditions to solver to determine if they
        # can satisfy the global constraint
        conditional_expressions = [cond.to_z3_expr() for (_, cond) in self.assertions]
        for expr in conditional_expressions:
            s.add(expr)

        if s.check() == z3.sat:
            # find a specific combination
            model = s.model()  # [cond1==True, cond2==False, cond3==False, ...]

            solver_with_conditional_constraint = z3.Solver()
            modelValues = {}
            # todo: launch multiple solvers in parallel, get first response

            # {var: makeTF(model[var]) for var in self.modelVariables}
            # assignment dict of condition:True/False: z3.z3.BoolRef
            self.assertions: List[Expression, Expression]
            for (conditional_constraint, condition) in self.assertions:
                # for unconditional conditions, or conditions that we assign to be true
                if condition == BoolVal(True) or model.eval(condition.to_z3_expr()):
                    # modelValues[condition] = BoolVal(True)
                    solver_with_conditional_constraint.add(conditional_constraint.to_z3_expr())
            solver_with_conditional_constraint.check()
            self.model = solver_with_conditional_constraint.model()
            return solver_with_conditional_constraint.check(*args)  # todo convert z3.sat to ours???? @sj, this would impact the current program using z3.sat tho

        else:
            raise "Impossible to find any way of building constraints"

    def model(self):
        return self.model()


sat = "sat"
unsat = "unsat"
timeout = "timeout"
unknown = "unknown"

def my_test():
    solver = Solver()
    # Adding unconditional constraints
    solver.add(Bool("x"), condition=BoolVal(True))  # Unconditionally true
    solver.add(Bool("y"), condition=BoolVal(True))  # Unconditionally true

    # Should be satisfiable without additional conditions
    result = solver.check()
    assert result==z3.sat


if __name__ == '__main__':
    s = Solver()
    # s.add(Or(Bool("x"), Bool("y")), Eq(Bool("x"), Bool("y")))
    s.add(Or(Bool("x"), Bool("y")))
    s.add(Implies(Bool("x"), Bool("y")), Bool("redundant_implies"))
    if s.check() == z3.sat:
        m = s.model
        print(m)
        # assert (m['x'])
        # assert (m['y'])
        print('test passed')
    else:
        print('something is very wrong')
    my_test()





# time_analysis.py
import os
import matplotlib.pyplot as plt

def read_time_from_line(file_path, line_num):
    """Read a specific line from a file and return the time value and timeout status."""
    with open(file_path, 'r') as file:
        for i, line in enumerate(file):
            if i == line_num:
                time_taken, timeout_status = line.strip().split(',')
                return float(time_taken), timeout_status == '1'
    return None, False


def escape_latex(s):
    """Escapes LaTeX special characters in a given string."""
    return s.replace('#', '\\#').replace('%', '\\%').replace('&', '\\&')


def generate_latex(only_constraint_false_timeout_count, both_constraint_timeout_count,
                   both_no_timeout_count, only_constraint_true_timeout_count,
                   constraint_true, constraint_false):
    # Escape special LaTeX characters in constraint names
    constraint_true = constraint_true.replace('_', '\\_')
    constraint_false = constraint_false.replace('_', '\\_')

    # calculate percentage:
    total_count = only_constraint_false_timeout_count + both_constraint_timeout_count + \
                  both_no_timeout_count + only_constraint_true_timeout_count

    # Format numbers as percentages with two decimal places and include the raw count
    only_constraint_false_timeout = f"{only_constraint_false_timeout_count} ({only_constraint_false_timeout_count / total_count * 100:.2f}\\%)"
    both_constraint_timeout = f"{both_constraint_timeout_count} ({both_constraint_timeout_count / total_count * 100:.2f}\\%)"
    both_no_timeout = f"{both_no_timeout_count} ({both_no_timeout_count / total_count * 100:.2f}\\%)"
    only_constraint_true_timeout = f"{only_constraint_true_timeout_count} ({only_constraint_true_timeout_count / total_count * 100:.2f}\\%)"

    # # Format numbers as percentages with two decimal places
    # only_constraint_false_timeout = f"{only_constraint_false_timeout * 100:.2f}\\%"
    # both_constraint_timeout = f"{both_constraint_timeout * 100:.2f}\\%"
    # both_no_timeout = f"{both_no_timeout * 100:.2f}\\%"
    # only_constraint_true_timeout = f"{only_constraint_true_timeout * 100:.2f}\\%"

    latex_code: str = f"""
\\begin{{table}}[ht]
\\centering
\\begin{{tabular}}{{|m{{1.5cm}}|c|c|c|}}
\\cline{{3-4}}
\\multicolumn{{2}}{{c|}}{{}} & \\multicolumn{{2}}{{c|}}{{{constraint_true}}} \\\\ \\cline{{3-4}} 
\\multicolumn{{2}}{{c|}}{{}} & No Timeout & Timeout \\\\ \\hline
\\multirow{{2}}{{*}}{{{constraint_false}}} & Timeout & {only_constraint_false_timeout} & {both_constraint_timeout} \\\\ \\cline{{2-4}} 
 & No Timeout & {both_no_timeout} & {only_constraint_true_timeout} \\\\ \\hline
\\end{{tabular}}
\\caption{{Comparison of timeout occurrences {constraint_true} and {constraint_false}}}
\\end{{table}}
"""
    return latex_code


def plot_comparison_for_constraint(files_directory, constraint_true='distinct', constraint_false='PbEq',
                                   time_cap=5,save_as_file:str=""):
    all_files = os.listdir(files_directory)
    times_full_true = []
    times_full_false = []
    times_holes_true = []
    times_holes_false = []
    timeout_only_false = 0
    timeout_both = 0
    never_timed_out = 0
    timeout_only_true = 0

    for file_name in all_files:
        if constraint_true in file_name:
            constraint_false_file = file_name.replace(constraint_true, constraint_false)
            constraint_false_file_path = os.path.join(files_directory, constraint_false_file)
            if not os.path.isfile(constraint_false_file_path):
                continue
            constraint_true_file_path = os.path.join(files_directory, file_name)
            time_type = 'full' if 'full_time' in file_name else 'holes'

            with open(constraint_true_file_path, 'r') as true_file, open(constraint_false_file_path, 'r') as false_file:
                for i, (true_line, false_line) in enumerate(zip(true_file, false_file)):
                    true_time, true_timeout = read_time_from_line(constraint_true_file_path, i)
                    false_time, false_timeout = read_time_from_line(constraint_false_file_path, i)

                    # classify all points that take more than time_cap as timeout
                    true_timeout = true_time > time_cap
                    false_timeout = false_time > time_cap

                    true_time = min(true_time, time_cap)
                    false_time = min(false_time, time_cap)
                    # Update timeout table
                    if true_timeout and not false_timeout:
                        timeout_only_true += 1
                    elif not true_timeout and false_timeout:
                        timeout_only_false += 1
                    elif true_timeout and false_timeout:
                        timeout_both += 1
                    else:
                        never_timed_out += 1

                    # Plot instances
                    if time_type == 'full':
                        times_full_true.append(true_time)
                        times_full_false.append(false_time)
                    else:
                        times_holes_true.append(true_time)
                        times_holes_false.append(false_time)

    # Plotting
    plt.figure(figsize=(6, 6))
    plt.scatter(times_full_true, times_full_false, color='green', alpha=0.5, label='Full Time')
    plt.scatter(times_holes_true, times_holes_false, color='blue', alpha=0.5, label='Holes Time')
    plt.plot([0, time_cap], [0, time_cap], 'r--')
    plt.xlim(0, time_cap)
    plt.ylim(0, time_cap)
    plt.xlabel(f'Times for {constraint_true} capped at {time_cap} seconds')
    plt.ylabel(f'Times for {constraint_false} capped at {time_cap} seconds')
    plt.title(f'Comparison of Times: {constraint_true} vs. {constraint_false} within [0, {time_cap}] seconds')
    plt.legend()
    plt.grid(True)
    if save_as_file:
        plt.savefig(save_as_file)
    plt.show()
    timeout_count_lst = [timeout_only_false, timeout_both, never_timed_out, timeout_only_true]

    print(generate_latex(*timeout_count_lst, constraint_true, constraint_false))
    return timeout_only_true, timeout_both, timeout_only_false, timeout_both


if __name__ == '__main__':
    TIME_OUT = 5

    files_directory = './time-record/whole_problem_time_records'
    # plot_comparison_for_constraint(files_directory, 'distinct', 'PbEq', time_cap=200)  # extended
    plot_name = "full_problem_no_num comparison"
    # plot_comparison_for_constraint(files_directory, 'is_num', 'is_bool',
    #                                time_cap=5,)
    constraint_true_false_lst = [
        ("classic","argyle"),
        ("distinct","PbEq"),
        ("percol","inorder"),
        ("is_bool","is_num"),
        ("prefill","no_prefill")
    ]
    # for (true_constraint,false_constraint) in constraint_true_false_lst:
    #     plot_comparison_for_constraint(files_directory, true_constraint,false_constraint,
    #                                    save_as_file='./images/whole_distinct_5s_cap', time_cap=5)

    plot_comparison_for_constraint(files_directory,"distinct","PbEq",save_as_file='./images/whole_distinct_5s_cap',time_cap=5)
    # forgot to handle timeout


import subprocess
import time
import os
from pathlib import Path
import warnings

class SMTFileErrorWarning(UserWarning):
    pass

def run_cvc5(smt2_file, time_out: int = 5):
    cvc_path = get_executable_path("cvc5-macOS-arm64")
    command = [cvc_path, smt2_file, "--lang", "smt2"]
    start_time = time.time()
    did_timeout = False
    try:
        result = subprocess.run(command,
                                capture_output=True, text=True, timeout=time_out)
        combined_output = ((result.stdout if result.stdout is not None else "") +
                           (result.stderr if result.stderr is not None else ""))  # capture all output
    except subprocess.TimeoutExpired as exc:
        did_timeout = True
        combined_output = ((exc.stdout.decode('utf-8') if exc.stdout else "") +
                           (exc.stderr.decode('utf-8') if exc.stderr else ""))  # capture all output
    ans = "timeout"

    end_time = time.time()
    total_time = end_time - start_time
    if not did_timeout:
        if "unsat" in combined_output:
            ans = "unsat"
        elif "sat" in combined_output:
            ans = "sat"
        elif "error" in combined_output.lower() or "unsupported" in combined_output.lower():
            ans = "error"
            warning_message = f"CVC5 encountered an error while processing {smt2_file}:\n{combined_output}"
            warnings.warn(warning_message, SMTFileErrorWarning)
        else:
            ans = "unknown"
    else:
        total_time = time_out
    return (total_time, did_timeout, ans)


def run_z3(smt2_file: str, time_out: int = 5):
    """
    :param smt_log_file_path:
    :param time_out: in seconds
    :return:
    """
    start_time = time.time()
    did_timeout = False
    try:
        result = subprocess.run(["z3", "-smt2", smt2_file],
                                capture_output=True, text=True, timeout=time_out)
        combined_output = ((result.stdout if result.stdout is not None else "") +
                           (result.stderr if result.stderr is not None else ""))  # capture all output
        print(f'z3 std output: {result.stdout}'
              f'z3 stderr output: {result.stderr}')
    except subprocess.TimeoutExpired as exc:
        did_timeout = True
        combined_output = ''
        result = exc
    ans = "timeout"
    end_time = time.time()
    total_time = end_time - start_time
    if not did_timeout:
        if "unsat" in combined_output:
            ans = "unsat"
        elif "sat" in combined_output:
            ans = "sat"
        elif "error" in combined_output.lower() or "unsupported" in combined_output.lower():
            ans = "error"
            warning_message = f"Z3 encountered an error while processing {smt2_file}:\n{combined_output}"
            warnings.warn(warning_message, SMTFileErrorWarning)
        else:
            ans = "unknown"
    else:
        total_time = time_out
    return total_time, did_timeout, ans


def run_yices(smt2_file):
    command = f"yices-smt2 {smt2_file}"
    try:
        result = subprocess.run(command, shell=True, capture_output=True, text=True, check=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        return f"Error: {e}"


# Dictionary to map solver names to their corresponding functions
solvers = {
    "cvc5": run_cvc5,
    "z3": run_z3
    # "yices": run_yices
}


def run_solvers(smt2_file:str='', smt2_str:str='', verbose=False, time_out=5):
    """
    time_out: in seconds
    """
    results = {}
    if smt2_str and smt2_file=='':
        smt2_file = os.path.join(os.path.dirname(__file__), 'smt_file.smt2')
        with open(smt2_file, 'w') as f:
            f.truncate()
            f.write(smt2_str)

    for solver, run_function in solvers.items():
        if verbose:
            print(f"Running {solver}...")
        result = run_function(smt2_file,time_out=time_out)
        results[solver] = result

    return results


def get_executable_path(solver_path_in_solvers_dir):
    # Get the directory of the current file (__file__ refers to the script in which this code is written)
    dir_of_jz3 = Path(os.path.dirname(__file__)).parent

    # Build the path to the executable
    executable_path = os.path.join(dir_of_jz3, "solvers/" + solver_path_in_solvers_dir)

    return executable_path


if __name__ == '__main__':
    print(run_cvc5(
        '/Users/jiazhenghao/Desktop/CodingProjects/jz3/jz3/problems_instances/particular_hard_instances_records/smt2_files/04_25_00_58_021714024682.151314'))



import jz3 as z3
# Create solver instance
s = z3.Solver()
# Define variables
x = z3.Int('x')
y =z3. Int('y')
# Add constraints
s.add(x > 2, y < 10, x + 2*y == 7)
if s.check() == z3.sat: # checking constraints sat
    print(s.model()) # [y = 0, x = 7]
else:
    print("No solution")



from setuptools import setup, find_packages


setup(
    name='jz3',
    version='0.1.8',
    packages=find_packages(),
    description='A simple wrapper for Z3 solver',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Sebastiaan Joosten, Zheng Robert Jia',
    author_email='jia00129@umn.edu',
    url='https://github.com/Robert-Jia00129/jz3',
    license='MIT',
    install_requires=[
        'z3-solver',
        'matplotlib',
    ],
    package_data={
        'jz3': ['solvers/*']
    },
    python_requires='>3.11',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
)


import os
import random
import errno

def shuffle_files(files_dir):
    """Shuffle all lines in each file in the specified directory."""
    for filename in os.listdir(files_dir):
        file_path = os.path.join(files_dir, filename)
        if os.path.isfile(file_path):
            with open(file_path, 'r') as file:
                lines = file.readlines()
            random.shuffle(lines)
            with open(file_path, 'w') as file:
                file.writelines(lines)

def remove_duplicates_files(files_dir):
    """Remove duplicate lines from each file in the specified directory."""
    for filename in os.listdir(files_dir):
        file_path = os.path.join(files_dir, filename)
        if os.path.isfile(file_path):
            with open(file_path, 'r') as file:
                lines = set(file.readlines())
            with open(file_path, 'w') as file:
                file.writelines(lines)

def shuffle_and_remove_duplicates_files(files_dir):
    """Remove duplicate lines and then shuffle the rest in each file in the specified directory."""
    for filename in os.listdir(files_dir):
        file_path = os.path.join(files_dir, filename)
        if os.path.isfile(file_path):
            with open(file_path, 'r') as file:
                lines = set(file.readlines())
            lines = list(lines)
            random.shuffle(lines)
            with open(file_path, 'w') as file:
                file.writelines(lines)


def clean_dir(file_dir_lst):
    """Empties the contents of files in each specified directory without removing the files."""
    for dir_path in file_dir_lst:
        if os.path.exists(dir_path):
            for filename in os.listdir(dir_path):
                file_path = os.path.join(dir_path, filename)
                if os.path.isfile(file_path):
                    open(file_path, 'w').close()  # Open in write mode and immediately close to clear contents

def auto_clean():
    """Cleans predefined directories."""
    # TODO
    directories_to_clean = ['/path/to/dir1', '/path/to/dir2']
    clean_dir(directories_to_clean)

def delete_dir(file_dir_lst):
    """Deletes each specified directory and its contents if it exists."""
    for dir_path in file_dir_lst:
        try:
            if os.path.exists(dir_path):
                for filename in os.listdir(dir_path):
                    file_path = os.path.join(dir_path, filename)
                    os.remove(file_path)
                os.rmdir(dir_path)
        except OSError as e:
            if e.errno != errno.ENOENT:  # No such file or directory
                raise  # Re-raise exception if a different error occurred

def delete_experiment():
    """Deletes predefined directories for a hard reset of experiments."""
    # TODO
    directories_to_delete = ['/path/to/experiment1', '/path/to/experiment2']
    delete_dir(directories_to_delete)


if __name__=='__main__':
    pass